#ifndef uterm_drm3d_blend_frag_H
#define uterm_drm3d_blend_frag_H
#include <stddef.h>
extern const char *const _binary_uterm_drm3d_blend_frag_start;
extern const char *const _binary_uterm_drm3d_blend_frag_end;
extern size_t _binary_uterm_drm3d_blend_frag_size;
#endif // uterm_drm3d_blend_frag_H