#ifndef uterm_input_fallback_xkb_H
#define uterm_input_fallback_xkb_H
#include <stddef.h>
extern const char *const _binary_uterm_input_fallback_xkb_start;
extern const char *const _binary_uterm_input_fallback_xkb_end;
extern size_t _binary_uterm_input_fallback_xkb_size;
#endif // uterm_input_fallback_xkb_H