#ifndef text_gltex_atlas_frag_H
#define text_gltex_atlas_frag_H
#include <stddef.h>
extern const char *const _binary_text_gltex_atlas_frag_start;
extern const char *const _binary_text_gltex_atlas_frag_end;
extern size_t _binary_text_gltex_atlas_frag_size;
#endif // text_gltex_atlas_frag_H